# Copyright (c) 2023 John Heintz, Gist Labs https://gistlabs.com
# License Apache v2 http://www.apache.org/licenses/
import numpy.random
from numpy.random import Generator, PCG64
import threading


class RandomLocal(threading.local):
    """
    A threadlocal to hold a stack of stable random generators
    """
    stack = []

    def push(self):
        """Push a new Generator PCG64 SeedSequence to the stack"""
        self.stack.append(Generator(PCG64(123456789)))

    def pop(self):
        """Removes the top element from the stack"""
        self.stack.pop(0)

    def top(self) -> numpy.random.Generator:
        """Lookup the top random generator, or return None

        Returns
        -------
        numpy.random.Generator or None
            Top random generator, or None on failure.
        """
        if len(self.stack) > 0:
            return self.stack[-1]
        else:
            return None


_randomLocalStack = RandomLocal()
_globalRandomGenerator = Generator(PCG64())


def random() -> Generator:
    """
    Return a Generator,either the current stable one or else a (random) global Generator

    This can be used for all cases to get random numbers, whether inside a @stablerandom or not

    """
    return _randomLocalStack.top() or _globalRandomGenerator


def stablerandom(func):
    """A decorator indicating stable random
    @stablerandom can be used stable random behavior for a function and call stack

    Usage:
        @stable_random
        def my_method():
            ...
    """
    def wrapper(*args, **kwargs):
        _randomLocalStack.push()

        try:
            return func(*args, **kwargs)
        finally:
            # clear thread local
            _randomLocalStack.pop()

    return wrapper


# random functions to wrap for stable,
# from https://numpy.org/doc/stable/reference/random/legacy.html#functions-in-numpy-random


_random_functions = ('normal', 'pareto', 'triangular', 'uniform')


def _wrap_numpy_random(funcName):
    _orig = getattr(numpy.random, funcName)

    def _stable(*args, **kwargs):
        stable = _randomLocalStack.top()
        if stable:
            func = getattr(stable, funcName)
            return func(*args, **kwargs)
        else:
            return _orig(*args, **kwargs)

    setattr(numpy.random, funcName, _stable)


[_wrap_numpy_random(f) for f in _random_functions]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stablerandom']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22,<2.0']

setup_kwargs = {
    'name': 'stablerandom',
    'version': '0.4.27',
    'description': 'Create stable/repeatable numpy.random applications',
    'long_description': '# stablerandom\nstablerandom provides a stable and repeatable implementation of \nthe NumPy random number generator (numpy.random). With this package, \nyou can generate the same sequence of random numbers across different \nplatforms and Python environments, ensuring reproducibility \nin scientific computing, machine learning, and unit testing.\n\nstablerandom can decorate any function or method and provide a \ncall-stack scoped seeded random generator. It is thread safe and \nsupports nested scopes.\n\n## Example\nUsing the `@stablerandom` decorator to get a stable output for numpy.random.triangular\n\n ```\nimport numpy.random\nfrom stablerandom import stablerandom\n\n@stablerandom\ndef random_triangular(samples):\n    return numpy.random.triangular(1, 5, 10, samples)\n\nprint(random_triangular(3))\n>>> [1.99882862 7.95097645 7.68974243]\nprint(random_triangular(3))\n>>> [1.99882862 7.95097645 7.68974243]\n```\n\n## Installing\n```bash\n$ pip install stablerandom\n```\nThe source code is currently hosted on GitHub at \nhttps://github.com/GistLabs/stablerandom\nand published in PyPI at https://pypi.org/project/stablerandom/ \n\nThe versioning scheme currently used is {major}.{minor}.{auto build number}\nfrom `git rev-list --count HEAD`. \n\nWe recommend picking a version like:\n\n`stablerandom = "^0.3"`\n\n## Dependencies\nThis library has been tested with [NumPy](https://www.numpy.org) back to version 1.22\n\n## Community guidelines\nWe welcome contributions and questions. Please head over to github and \nsend us pull requests or create issues!\n',
    'author': 'John Heintz',
    'author_email': 'john@gistlabs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
